unit dnPxMask;

(*****************************************************************************
Copyright  1997 Dieter Nth 
version    : 1.02
date       : 19.07.1997
description:

  This unit implements Paradox-style Inputmasks:
  most of the validation-code of Inputmasks is based upon Borland's VALIDATE.PAS
  Unit ( 1992 Borland International) from Borland Pascal 7, 
  i don't know why they dropped it from VCL...

  most of the TPxCustomMaskEdit-class is based upon code from TCustomMaskEdit

Bugs fixed:
1.02 :
  Borland's example input mask (from DBD)
  *[&[*?]*[{#, ,.,;,,;;,:,;[,;],(,),}]]  (capitalized words) allowed only
  ONE word to be entered.
  Yes, I should have tried at least the example masks, but I did it once and
  then I found a bug in Borland's Iteration function causing infinite loops in
  iterations (without a number) containing groups....
  I thought, I fixed it, but this fix caused this new problem...
  There still may be some combinations of groups and iterations that will
  cause an infinite loop. So, if your running your program from within Delphi
  and it hangs while typing in a TPxMaskEdit, set a breakpoint in
  function Iteration and stop your program. If this happens within the
  PxMask-property editor you'll have to kill Delphi ;-).
  Before You send me a email complaining about this, please try to follow
  all those recursive calls with debugger (and remember, that part of code
  wasn't written by Borland, not me), maybe YOU'll fix it. If You don't,
  I'll try my best.

  Testing an input mask in DBD32 I noticed, that there's a new special char:
  ---> ~ <--- lower case Char
  and, of course, I implemented it.
  If You don't want to use this new feature, you may comment out that part in
  function Scan
******************************************************************************)

interface

uses
  {$IFDEF Ver80} WinTypes, WinProcs, {$ELSE} Windows, {$ENDIF}
  Messages, SysUtils, Classes, Controls, StdCtrls, Forms, Menus, Graphics;

type
  EPxMaskError = class(Exception);

  TCustomPxMaskEdit = class(TCustomEdit)
  private

    {$IFDEF Ver80}
    FMask: pString;
    FOldValue: pString;
    {$ELSE}
    FMask: String;
    FOldValue: String;
    {$ENDIF}

    FErrPos : integer;
    FReEnter,
    FAutoValidate,
    FAutoFill,
    FRaiseInputError : boolean;

    procedure setMask(const value : string);
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;

    {$IFDEF Ver80}
    function getMask : string;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    {$ELSE}
    procedure CMWantSpecialKey(var Message: TCMWantSpecialKey); message CM_WANTSPECIALKEY;
    {$ENDIF}

  protected
    function EditCanModify: Boolean; virtual;
    procedure KeyPress(var Key: Char); override;
    function isMasked : boolean;
    {$IFDEF Ver80}
    property PxMask : string read GetMask write SetMask;
    {$ELSE}
    property PxMask : string read FMask write SetMask;
    {$ENDIF}
  public
    constructor create(aOwner : TComponent); override;

    {$IFDEF Ver80}
    destructor Destroy; override;
    {$ENDIF}

    procedure CheckSyntax(const aMask : string);
    procedure Reset; virtual;
    procedure ValidateEdit; virtual;
    procedure ValidateError; dynamic;

    {*** true: performs ValidateEdit on exiting the control}
    property AutoValidate : boolean read FAutoValidate write FAutoValidate default true;

    {*** true : Input will automatically be completed}
    property AutoFill : boolean read FAutoFill write FAutoFill default true;

    {*** Position of last Error after any validation}
    property ErrorPos : integer read FErrPos;

    {*** true : if typing an illegal char, exception is raised, so you can catch it and
         redirect to a statuspanel}
    property RaiseInputError : boolean read FRaiseInputError
                                       write FRaiseInputError
                                       default false;
  published
  end;

  TPxMaskEdit = class(TCustomPxMaskEdit)
  published
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    {$IFNDEF Ver80} property OnStartDrag; {$ENDIF}
    property AutoFill;
    property AutoValidate;
    property ErrorPos;
    property PxMask;
    property RaiseInputError;
  end;

{*** check syntax of supplied mask, if an exception is raised then ErrorPos > 0 }
procedure PxMaskCheckSyntax(const aMask: string; var ErrorPos: integer);

{*** aMask : ParadoxMask (no syntax checks)
     Input : string to be checked
     AutoFill : if true then Input will automatically be completed
     returns position of error in Input, if any: 0 = OK}
function PxMaskCheckInput(const aMask: string; var Input: string;
  AutoFill: Boolean): integer;

{*** aMask : ParadoxMask (no syntax checks)
     Input : string to be checked
     returns position of error in Input, or length(Input), if not complete}
function PxMaskValidate(const aMask: string; const Input: string): integer;

implementation

uses Mask,
     dnPxCnst;

{*** copied from VALIDATE.PAS (and partially modified)}
type
  TPxMaskResult = (prComplete, prIncomplete, prEmpty, prError, prSyntax,
    prAmbiguous, prIncompNoFill);

const
  SpecialChars = ['#', '?', '&', '!', '@', '*', '{', '}', '[', ']', ','];
  Digits = ['0' .. '9'];

function IsComplete(Rslt: TPxMaskResult): Boolean;
begin
  result := Rslt in [prComplete, prAmbiguous];
end;

function IsIncomplete(Rslt: TPxMaskResult): Boolean;
begin
  result := Rslt in [prIncomplete, prIncompNoFill];
end;

function UpCase(c : char) : char;
begin
  result := char(longInt(ansiUpper(pChar(longInt(c)))));
end;

function LoCase(c : char) : char;
begin
  result := char(longInt(ansiLower(pChar(longInt(c)))));
end;

function Picture(const aMask: string; var Input: string;
  AutoFill: Boolean; var ErrorPos : integer): TPxMaskResult;
var
  I, J: cardinal;
  Rslt: TPxMaskResult;
  Reprocess: Boolean;

  function Process(TermCh: cardinal): TPxMaskResult;
  var
    Rslt: TPxMaskResult;
    Incomp: Boolean;
    OldI, OldJ, IncompJ, IncompI: cardinal;

    { Consume input }

    procedure Consume(Ch: Char);
    begin
      Input[J] := Ch;
      Inc(J);
      Inc(I);
      inc(ErrorPos);
    end;

    { Skip a character or a picture group }

    procedure ToGroupEnd(var I: cardinal);
    var
      BrkLevel, BrcLevel: Integer;
    begin
      BrkLevel := 0;
      BrcLevel := 0;
      repeat
        if I = TermCh then Exit;
        case aMask[I] of
          '[': Inc(BrkLevel);
          ']': Dec(BrkLevel);
          '{': Inc(BrcLevel);
          '}': Dec(BrcLevel);
          ';': Inc(I);
          '*':
            begin
              Inc(I);
              while aMask[I] in Digits do Inc(I);
              ToGroupEnd(I);
              Continue;
            end;
        end;
        Inc(I);
      until (BrkLevel = 0) and (BrcLevel = 0);
    end;

    { Find the a comma separator }

    function SkipToComma: Boolean;
    begin
      repeat ToGroupEnd(I) until (I = TermCh) or (aMask[I] = ',');
      if aMask[I] = ',' then Inc(I);
      SkipToComma := I < TermCh;
    end;

    { Calclate the end of a group }

    function CalcTerm: cardinal;
    var
      K : cardinal;
    begin
      K := I;
      ToGroupEnd(K);
      CalcTerm := K;
    end;

    { The next group is repeated X times }

    function Iteration: TPxMaskResult;
    var
      Itr, K, L: cardinal;
      Rslt : TPxMaskResult;
      NewTermCh: cardinal;
    begin
      Itr := 0;
      Iteration := prError;

      Inc(I);  { Skip '*' }

      { Retrieve number }

      while aMask[I] in Digits do
      begin
        Itr := Itr * 10 + Byte(aMask[I]) - Byte('0');
        Inc(I);
      end;

      if I > TermCh then
      begin
        Iteration := prSyntax;
        Exit;
      end;

      K := I;
      NewTermCh := CalcTerm;

      { If Itr is 0 allow any number, otherwise enforce the number }
      if Itr <> 0 then
      begin
        for L := 1 to Itr do
        begin
          I := K;
          Rslt := Process(NewTermCh);
          if not IsComplete(Rslt) then
          begin
            { Empty means incomplete since all are required }
            if Rslt = prEmpty then Rslt := prIncomplete;
            Iteration := Rslt;
            Exit;
          end;
        end;
      end
      else
      begin
        repeat
          I := K;
          Rslt := Process(NewTermCh);
          {Fixes (hope so) a bug from Borland causing infinite loops}
          if (Rslt = prAmbiguous) and (I = NewTermCh) then
          begin
            Rslt := prComplete;
            break;
          end;
          {End Fix}
        until not IsComplete(Rslt);
        if (Rslt = prEmpty) or (Rslt = prError) then
        begin
          Inc(I);
          Rslt := prAmbiguous;
        end;
      end;
      I := NewTermCh;
      Iteration := Rslt;
    end;

    { Process a picture group }

    function Group: TPxMaskResult;
    var
      Rslt: TPxMaskResult;
      TermCh: cardinal;
    begin
      TermCh := CalcTerm;
      Inc(I);
      Rslt := Process(TermCh - 1);
      if not IsIncomplete(Rslt) then I := TermCh;
      Group := Rslt;
    end;

    function CheckComplete(Rslt: TPxMaskResult): TPxMaskResult;
    var
      J: cardinal;
    begin
      J := I;
      if IsIncomplete(Rslt) then
      begin
        { Skip optional pieces }
        while True do
          case aMask[J] of
            '[': ToGroupEnd(J);
            '*':
              if not (aMask[J + 1] in Digits) then
              begin
                Inc(J);
                ToGroupEnd(J);
              end
              else
                Break;
          else
            Break;
          end;

        if J = TermCh then Rslt := prAmbiguous;
      end;
      CheckComplete := Rslt;
    end;

    function Scan: TPxMaskResult;
    var
      Ch: Char;
      Rslt: TPxMaskResult;
    begin
      Scan := prError;
      Rslt := prEmpty;
      while (I <> TermCh) and (aMask[I] <> ',') do
      begin
        if J > Length(Input) then
        begin
          Scan := CheckComplete(Rslt);
          Exit;
        end;

        Ch := Input[J];
        case aMask[I] of
          '#': if not (ch in Digits) then Exit
               else Consume(Ch);
          '?': if not isCharAlpha(ch) then Exit
               else Consume(Ch);
          '&': if not isCharAlpha(ch) then Exit
               else Consume(UpCase(ch));
          {New special char in DBD32}
          '~': if not isCharAlpha(ch) then Exit
               else Consume(LoCase(ch));
          '!': Consume(UpCase(ch));
          '@': Consume(Ch);
          '*':
            begin
              Rslt := Iteration;
              if not IsComplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
              if Rslt = prError then Rslt := prAmbiguous;
            end;
          '{':
            begin
              Rslt := Group;
              if not IsComplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
            end;
          '[':
            begin
              Rslt := Group;
              if IsIncomplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
              if Rslt = prError then Rslt := prAmbiguous;
            end;
        else
          if aMask[I] = ';' then Inc(I);
         if (UpCase(aMask[i]) = UpCase(ch)) or (Ch = ' ')
          then Consume(aMask[I])
          else Exit;
        end;

        if Rslt = prAmbiguous then
          Rslt := prIncompNoFill
        else
          Rslt := prIncomplete;
      end;

      if Rslt = prIncompNoFill then
        Scan := prAmbiguous
      else
        Scan := prComplete;
    end;

  begin
    Incomp := False;
    OldI := I;
    OldJ := J;
    repeat
      Rslt := Scan;

      { Only accept completes if they make it farther in the input
        stream from the last incomplete }
      if (Rslt in [prComplete, prAmbiguous]) and Incomp and (J < IncompJ) then {???}
      begin
        Rslt := prIncomplete;
        J := IncompJ;
      end;

      if (Rslt = prError) or (Rslt = prIncomplete) then
      begin
        Process := Rslt;
        if not Incomp and (Rslt = prIncomplete) then
        begin
          Incomp := True;
          IncompI := I;
          IncompJ := J;
        end;
        I := OldI;
        J := OldJ;
        if not SkipToComma then
        begin
          if Incomp then
          begin
            Process := prIncomplete;
            I := IncompI;
            J := IncompJ;
          end;
          Exit;
        end;
        OldI := I;
      end;
    until (Rslt <> prError) and (Rslt <> prIncomplete);

    if (Rslt = prComplete) and Incomp then
      Process := prAmbiguous
    else
      Process := Rslt;
  end;

begin
  ErrorPos := 1;
  Picture := prComplete;
  if aMask = '' then exit;

  Picture := prEmpty;
  if Input = '' then Exit;

  J := 1;
  I := 1;

  Rslt := Process(Length(aMask) + 1);
  if (Rslt <> prError) and (Rslt <> prSyntax) and (J <= Length(Input)) then
    Rslt := prError;

  if (Rslt = prIncomplete) and AutoFill then
  begin
    Reprocess := False;
    while (I <= Length(aMask)) and
      not (aMask[I] in SpecialChars) do
    begin
      if aMask[I] = ';' then Inc(I);
      Input := Input + aMask[I];
      Inc(I);
      Reprocess := True;
    end;
    J := 1;
    I := 1;
    if Reprocess then
      Rslt := Process(Length(aMask) + 1)
  end;

  case Rslt of
    prAmbiguous : Picture := prComplete;
    prIncompNoFill : Picture := prIncomplete;
  else
    Picture := Rslt;
  end;
end;


procedure PxMaskCheckSyntax(const aMask : string; var ErrorPos : integer);
const
  MaxBracketLevel = 32;  { maximale Anzahl der Klammerebenen }
var
  i: Integer;
  fStack : array[0..pred(MaxBracketLevel)] of char; {Klammerstack}
  fBracketLevel : shortInt;

  procedure RaiseError(const Err : string);
  begin
    ErrorPos := i;
    raise EPxMaskError.create(LoadStr(SPxMaskSyntax)+ #13#10 + Err);
  end;

  procedure PopBracket;
  begin
    if fBracketLevel > 0 then
    begin
      dec(fBracketLevel);
      if (fStack[fBracketLevel] <> aMask[i]) then
        RaiseError(FmtLoadStr(SPxMaskBracketError, [fStack[fBracketLevel]]));
    end else
      raiseError(FmtLoadStr(SPxMaskBracketClose, [aMask[i]]));
  end;

  procedure PushBracket(c : char);
  begin
    if fBracketLevel >= MaxBracketLevel then
      raiseError(LoadStr(SPxMaskBracketLevel));
    fStack[fBracketLevel] := c;
    inc(fBracketLevel);
  end;

begin
  if aMask = '' then Exit;
  fBracketLevel := 0;
  ErrorPos := 0;
  i := 1;
  while i <= Length(aMask) do
  begin
    case aMask[i] of
      '[': pushBracket(']');
      '{': pushBracket('}');
      ']', '}':
        begin
          popBracket;
          case aMask[i-1] of
            '[', '{':
              if (i <= 2) or (aMask[i - 2] <> ';') then
                raiseError(LoadStr(SPxMaskBracketEmpty));
          end;
        end;
      ';': Inc(i);
      '*':
        begin
          inc(i);
          while aMask[i] in Digits do inc(i);
          if (i > length(aMask)) or (aMask[i] in [']', '}', '*']) then
            RaiseError(LoadStr(SPxMaskIterationExpected));
          continue;
        end;
    end;
    Inc(i);
  end;
  if fBracketLevel > 0 then
  begin
    dec(fBracketLevel);
    RaiseError(FmtLoadStr(SPxMaskBracketMissing, [fStack[fBracketLevel]]));
  end;
  i := length(aMask);
  if (aMask[i] = ';') and (aMask[i - 1] <> ';') then
    RaiseError(LoadStr(SPxMaskLiteralExpected));
end;

function PxMaskCheckInput(const aMask   : string;
                            var   Input   : string;
                                  AutoFill: Boolean): integer;
begin
  if not (Picture(aMask, Input, AutoFill, result) in [prError, prSyntax]) then
    result := 0;
end;

function PxMaskValidate(const aMask : string;
                          const Input : string): integer;
var
  Str: String;
begin
  Str := Input;
  if Picture(aMask, Str, False, result) in [prComplete, prEmpty] then
    result := 0;
end;

{**** TCustomPxMaskEdit ****}

constructor TCustomPxMaskEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  {$IFDEF Ver80}
  FMask := NullStr;
  FOldValue := NullStr;
  {$ENDIF}

  FAutoFill := true;
  FAutoValidate := true;
  FRaiseInputError := false;
end;

{$IFDEF Ver80}
destructor TCustomPxMaskEdit.Destroy;
begin
  disposeStr(FMask);
  disposeStr(fOldValue);
  inherited Destroy;
end;
{$ENDIF}

procedure TCustomPxMaskEdit.Reset;
begin
  if Modified then
  begin
    {$IFDEF Ver80} Text := FOldValue^; {$ELSE} Text := FOldValue; {$ENDIF}
    Modified := False;
  end;
end;

procedure TCustomPxMaskEdit.setMask(const value : string);
begin
  if value <> PxMask then
  begin
    if value <> '' then CheckSyntax(value);
    {$IFDEF Ver80} assignStr(FMask, value); {$ELSE} FMask := value; {$ENDIF}
  end;
end;

function TCustomPxMaskEdit.isMasked : boolean;
begin
  {$IFDEF Ver80}
  result := FMask <> NullStr;
  {$ELSE}
  result := FMask <> '';
  {$ENDIF}
end;

{$IFDEF Ver80}
function TCustomPxMaskEdit.getMask : string;
begin
  result := FMask^;
end;
{$ENDIF}

procedure TCustomPxMaskEdit.KeyPress(var Key: Char);
var
  s : string;
begin
  inherited KeyPress(Key);
  if cardinal(key) = VK_ESCAPE then
  begin
    Reset;
    key := #0;
    exit;
  end;

  if IsMasked then
  begin
    if (Key in [#32..#255]) then
    begin
      if (not EditCanModify) or ReadOnly then
        key := #0
      else
      begin
        s := copy(text, 1, selStart) + key;
        if ((selStart + selLength) = length(text)) then
        begin
          Key := #0;
          FErrPos := PxMaskCheckInput(PxMask, s, AutoFill);
          if ErrorPos = 0 then
          begin
            selectAll;
            {$IFDEF Ver80}
            sendMessage(Handle, EM_REPLACESEL, 0, LongInt(StrPCopy(@s, s)));
            {$ELSE}
            SendMessage(Handle, EM_REPLACESEL, 0, LongInt(PChar(s)));
            {$ENDIF}

          end;
        end
        else
        begin
          FErrPos := PxMaskCheckInput(PxMask, s, false);
          if ErrorPos = 0 then key := s[length(s)] else key := #0;
        end;
        if ErrorPos = 0 then
          modified := true
        else
        begin
          selStart := ErrorPos - 1;
          messageBeep($FFFF);
          if RaiseInputError then
            raise EDBEditError.createRes(SPxMaskIllegalChar);
        end;
      end;
    end
    else
      if cardinal(key) = VK_RETURN then
      begin
        validateEdit;
        exit;
      end;
  end;
end;

function TCustomPxMaskEdit.EditCanModify: Boolean;
begin
  Result := true;
end;

procedure TCustomPxMaskEdit.ValidateEdit;
begin
  if IsMasked and modified then
  begin
    FErrPos := PxMaskValidate(PxMask, Text);
    if (FErrPos <> 0) then
    begin
      if not (csDesigning in ComponentState)then
     begin
       FReEnter := true;
       SetFocus;
     end;
     ValidateError;
    end;
  end;
end;

procedure TCustomPxMaskEdit.ValidateError;
begin
  messageBeep(0);
  raise EDBEditError.createRes(SPxMaskIllegalString);
end;

procedure TCustomPxMaskEdit.CMTextChanged(var Message: TMessage);
begin
  inherited;
  {$IFDEF Ver80} AssignStr(FOldValue, Text); {$ELSE} FOldValue := Text; {$ENDIF}
end;

{$IFDEF Ver80}
procedure TCustomPxMaskEdit.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  if IsMasked and Modified then  {will want to see escape key}
    Message.Result := Message.Result or DLGC_WANTALLKEYS;
end;
{$ELSE}
procedure TCustomPxMaskEdit.CMWantSpecialKey(var Message: TCMWantSpecialKey);
begin
  inherited;
  if (Message.CharCode = VK_ESCAPE) and IsMasked and Modified then
    Message.Result := 1;
end;
{$ENDIF}

procedure TCustomPxMaskEdit.CMExit(var Message: TCMExit);
begin
  if IsMasked and AutoValidate and not (csDesigning in ComponentState) then
    ValidateEdit;
  inherited;
end;

procedure TCustomPxMaskEdit.CMEnter(var Message: TCMExit);
begin
  if IsMasked and not (csDesigning in ComponentState) then
  begin
    if not FReEnter then
    begin
      {$IFDEF Ver80}
      AssignStr(FOldValue, Text);
      {$ELSE}
      FOldValue := Text;
      {$ENDIF}
      inherited;
    end
    else
    begin
      SelStart := FErrPos;
      FReEnter := false;
    end
  end
  else
    inherited;
end;

procedure TCustomPxMaskEdit.CheckSyntax(const aMask : string);
begin
  PxMaskCheckSyntax(aMask, FErrPos);
end;

end.
